@@ -95,7 +95,7 @@ class GoogleDriveDownloader:
         self._class_logger.debug('Initializing GoogleDriveDownloader')
         self._class_logger.debug('Output directory: %s', out_dir)
         self.out_dir = Path(out_dir)
-        regex_pattern = 'https:\/\/drive.google.com\/file\/d\/(\S*)\/.*'  # pylint:disable=anomalous-backslash-in-string
+        regex_pattern = '(\/|=)([-\w]{25,})'  # pylint:disable=anomalous-backslash-in-string
         self._class_logger.debug('Regex pattern: %s', regex_pattern)
         self.regex = re.compile(regex_pattern)
 
@@ -178,24 +178,76 @@ class GoogleDriveDownloader:
         match = self.regex.search(sharing_link)
         if match:
             try:
-                return match.group(1)
+                return match.group(2)
             #: Not sure how this would happen (can't even figure out a test), but leaving in for safety.
             except IndexError as err:
                 raise IndexError(f'Regex could not extract the file id from sharing link {sharing_link}') from err
         raise RuntimeError(f'Regex could not match sharing link {sharing_link}')
 
-    def download_image_from_google_drive(self, sharing_link):
+    def download_file_from_google_drive(self, sharing_link, join_id):
         """Download a publicly-shared image from Google Drive using it's sharing link
 
         Args:
             sharing_link (str): The publicly-shared link to the image.
+            join_id (str or int): The unique key for the row (used for reporting)
+
+        Returns:
+            Path: Path of downloaded file or None if download fails/is not possible
         """
 
-        file_id = self._get_file_id_from_sharing_link(sharing_link)
-        self._class_logger.debug('Downloading file id %s', file_id)
-        response = self._get_http_response(file_id)
-        filename = self._get_filename_from_response(response)
-        self._save_response_content(response, self.out_dir / filename)
+        if not sharing_link:
+            self._class_logger.debug('Row %s has no attachment info', join_id)
+            return None
+        self._class_logger.debug('Row %s: downloading shared file %s', join_id, sharing_link)
+        try:
+            file_id = self._get_file_id_from_sharing_link(sharing_link)
+            self._class_logger.debug('Row %s: extracted file id %s', join_id, file_id)
+            response = self._get_http_response(file_id)
+            filename = self._get_filename_from_response(response)
+            out_file_path = self.out_dir / filename
+            self._save_response_content(response, out_file_path)
+            return out_file_path
+        except Exception as err:
+            self._class_logger.warning('Row %s: Couldn\'t download %s', join_id, sharing_link)
+            self._class_logger.warning(err)
+            return None
+
+    #: TODO: WIP
+    def download_file_from_google_drive_using_api(self, sharing_link, join_id):
+        if not sharing_link:
+            self._class_logger.debug('Row %s has no attachment info', join_id)
+            return None
+        self._class_logger.debug('Row %s: downloading shared file %s', join_id, sharing_link)
+        try:
+            file_id = self._get_file_id_from_sharing_link(sharing_link)
+            self._class_logger.debug('Row %s: extracted file id %s', join_id, file_id)
+            response = self._get_http_response(file_id)
+            filename = self._get_filename_from_response(response)
+            out_file_path = self.out_dir / filename
+            self._save_response_content(response, out_file_path)
+            return out_file_path
+        except Exception as err:
+            self._class_logger.warning('Row %s: Couldn\'t download %s', join_id, sharing_link)
+            self._class_logger.warning(err)
+            return None
+
+    def download_attachments_from_dataframe(self, dataframe, sharing_link_column, join_id_column, output_path_column):
+        """Download the attachments linked in a dataframe column, creating a new column with the resulting path
+
+        Args:
+            dataframe (pd.DataFrame): Input dataframe with required columns
+            sharing_link_column (str): Column holding the Google sharing link
+            join_id_column (str): Column holding a unique key (for reporting purposes)
+            output_path_column (str): Column for the resulting path; will be added if it doesn't existing in the dataframe
+
+        Returns:
+            pd.DataFrame: Input dataframe with output path info
+        """
+
+        dataframe[output_path_column] = dataframe.apply(
+            lambda x: self.download_file_from_google_drive(x[sharing_link_column], x[join_id_column]), axis=1
+        )
+        return dataframe
 
 
 class SFTPLoader:
