@@ -59,6 +59,19 @@ class FeatureServiceInlineUpdater:
 
         return rows_updated
 
+    def _validate_working_fields_in_live_and_new_dataframes(self, live_dataframe, fields):
+        live_fields = set(live_dataframe.columns)
+        new_fields = set(self.new_dataframe.columns)
+        working_fields = set(fields)
+
+        live_dif = working_fields - live_fields
+        new_dif = working_fields - new_fields
+
+        if live_dif or new_dif:
+            raise RuntimeError(
+                f'Field mismatch between defined fields and either new or live data.\nFields not in live data: {live_dif}\nFields not in new data: {new_dif}'
+            )
+
     def _get_common_rows(self, live_dataframe) -> pd.DataFrame:
         """Create a dataframe containing only the rows common to both the existing live dataset and the new updates
 
@@ -196,6 +209,7 @@ class FeatureServiceInlineUpdater:
         feature_layer_item = self.gis.content.get(feature_layer_itemid)
         feature_layer = arcgis.features.FeatureLayer.fromitem(feature_layer_item)
         live_dataframe = pd.DataFrame.spatial.from_layer(feature_layer)
+        self._validate_working_fields_in_live_and_new_dataframes(live_dataframe, fields)
         subset_dataframe = self._get_common_rows(live_dataframe)
         if subset_dataframe.empty:
             self._class_logger.warning(
@@ -220,6 +234,7 @@ class FeatureServiceAttachmentsUpdater:
     def __init__(self, gis):
         self.gis = gis
         self._class_logger = logging.getLogger(__name__).getChild(self.__class__.__name__)
+        self.failed_dict = {}
 
     #: This isn't used anymore... but it feels like a shame to lose it.
     @staticmethod
@@ -335,11 +350,17 @@ class FeatureServiceAttachmentsUpdater:
             filepath = row[attachment_path_field]
 
             self._class_logger.debug('Add %s to OID %s', filepath, target_oid)
-            result = self.feature_layer.attachments.add(target_oid, filepath)
-            self._class_logger.debug('%s', result)
+            try:
+                result = self.feature_layer.attachments.add(target_oid, filepath)
+            except Exception:
+                self._class_logger.error('AGOL error while adding %s to OID %s', filepath, target_oid, exc_info=True)
+                self.failed_dict[target_oid] = ('add', filepath)
+                continue
 
+            self._class_logger.debug('%s', result)
             if not result['addAttachmentResult']['success']:
                 warnings.warn(f'Failed to attach {filepath} to OID {target_oid}')
+                self.failed_dict[target_oid] = ('add', filepath)
                 continue
 
             adds_count += 1
@@ -370,19 +391,38 @@ class FeatureServiceAttachmentsUpdater:
             self._class_logger.debug(
                 'Overwriting %s (attachment ID %s) on OID %s with %s', old_name, attachment_id, target_oid, filepath
             )
-            result = self.feature_layer.attachments.update(target_oid, attachment_id, filepath)
-            self._class_logger.debug('%s', result)
+            try:
+                result = self.feature_layer.attachments.update(target_oid, attachment_id, filepath)
+            except Exception:
+                self._class_logger.error(
+                    'AGOL error while overwritting %s (attachment ID %s) on OID %s with %s',
+                    old_name,
+                    attachment_id,
+                    target_oid,
+                    filepath,
+                    exc_info=True
+                )
+                self.failed_dict[target_oid] = ('update', filepath)
+                continue
 
             if not result['updateAttachmentResult']['success']:
                 warnings.warn(
                     f'Failed to update {old_name}, attachment ID {attachment_id}, on OID {target_oid} with {filepath}'
                 )
+                self.failed_dict[target_oid] = ('update', filepath)
                 continue
 
             overwrites_count += 1
 
         return overwrites_count
 
+    @staticmethod
+    def _check_attachment_dataframe_for_invalid_column_names(attachment_dataframe, invalid_names):
+        invalid_names_index = pd.Index(invalid_names)
+        intersection = attachment_dataframe.columns.intersection(invalid_names_index)
+        if not intersection.empty:
+            raise RuntimeError(f'Attachment dataframe contains the following invalid names: {list(intersection)}')
+
     def update_attachments(
         self, feature_layer_itemid, attachment_join_field, attachment_path_field, attachments_df, layer_number=0
     ):
@@ -407,12 +447,20 @@ class FeatureServiceAttachmentsUpdater:
         """
 
         self._class_logger.info('Updating attachments...')
+        #: These names are present in the live attachment data downloaded from AGOL. Because we merge the dataframes
+        #: later, we need to make sure they're not the same. There may be better ways of handling this that allows the
+        #: client names to be preserved, but for now force them to fix this.
+        self._check_attachment_dataframe_for_invalid_column_names(
+            attachments_df, invalid_names=['OBJECTID', 'PARENTOBJECTID', 'NAME', 'ID']
+        )
         self._class_logger.debug('Using layer %s from item ID %s', layer_number, feature_layer_itemid)
         self.feature_layer = self.gis.content.get(feature_layer_itemid).layers[layer_number]
         live_features_as_df = pd.DataFrame.spatial.from_layer(self.feature_layer)
         live_data_subset_df = self._get_live_oid_and_guid_from_join_field_values(
             live_features_as_df, attachment_join_field, attachments_df
         )
+        #: TODO: Make sure layer supports attachments so we don't get an arcgis error.
+        #: Check out the feature layer .properties and FeatureLayerManager.add_to_definition to check/enable?
         attachment_eval_df = self._get_current_attachment_info_by_oid(live_data_subset_df)
         attachment_action_df = self._create_attachment_action_df(attachment_eval_df, attachment_path_field)
 
@@ -422,6 +470,30 @@ class FeatureServiceAttachmentsUpdater:
 
         return overwrites_count, adds_count
 
+    @staticmethod
+    def build_attachments_dataframe(input_dataframe, join_column, attachment_column, out_dir):
+        """Create an attachments dataframe by subsetting down to just the two fields and dropping any rows
+           with null/empty attachments
+
+        Args:
+            input_dataframe (pd.DataFrame): Input data containing at least the join and attachment filename columns
+            join_column (str): Unique key joining attachments to live data
+            attachment_column (str): Filename for each attachment
+            out_dir (str or Path): Output directory, will be used to build full path to attachment
+
+        Returns:
+            pd.DataFrame: Dataframe with join key, attachment name, and full attachment paths
+        """
+
+        input_dataframe[attachment_column].replace('', np.nan, inplace=True)  #: pandas doesn't see empty strings as NAs
+        attachments_dataframe = input_dataframe[[join_column, attachment_column]] \
+                                            .copy().dropna(subset=[attachment_column])
+        #: Create the full path by prepending the output directory using .apply and a lambda function
+        attachments_dataframe['full_file_path'] = attachments_dataframe[attachment_column] \
+                                                    .apply(lambda filename: str(Path(out_dir, filename)))
+
+        return attachments_dataframe
+
 
 class FeatureServiceOverwriter:
     """Overwrites an AGOL Feature Service with data from a pandas DataFrame and a geometry source (Spatially-enabled
