@@ -1,6 +1,7 @@
 import builtins
 import json
 import logging
+from pathlib import Path
 
 import arcgis
 import numpy as np
@@ -783,6 +784,70 @@ class TestFeatureServiceInlineUpdaterResultParsing:
         }
 
 
+class TestFeatureServiceInlineUpdaterFieldValidation:
+
+    def test_validate_working_fields_in_live_and_new_dataframes_doesnt_raise_when_matching(self, mocker):
+        live_df = pd.DataFrame(columns=['field1', 'field2', 'field3'])
+        new_df = pd.DataFrame(columns=['field1', 'field2', 'field3'])
+        fields = ['field1', 'field2', 'field3']
+
+        updater_mock = mocker.Mock()
+        updater_mock.new_dataframe = new_df
+
+        #: This shouldn't raise an exception and so the test should pass
+        palletjack.FeatureServiceInlineUpdater._validate_working_fields_in_live_and_new_dataframes(
+            updater_mock, live_df, fields
+        )
+
+    def test_validate_working_fields_in_live_and_new_dataframes_raises_not_in_live(self, mocker):
+        live_df = pd.DataFrame(columns=['field1', 'field2'])
+        new_df = pd.DataFrame(columns=['field1', 'field2', 'field3'])
+        fields = ['field1', 'field2', 'field3']
+
+        updater_mock = mocker.Mock()
+        updater_mock.new_dataframe = new_df
+
+        with pytest.raises(RuntimeError) as exc_info:
+            palletjack.FeatureServiceInlineUpdater._validate_working_fields_in_live_and_new_dataframes(
+                updater_mock, live_df, fields
+            )
+        assert exc_info.value.args[
+            0
+        ] == 'Field mismatch between defined fields and either new or live data.\nFields not in live data: {\'field3\'}\nFields not in new data: set()'
+
+    def test_validate_working_fields_in_live_and_new_dataframes_raises_not_in_new(self, mocker):
+        live_df = pd.DataFrame(columns=['field1', 'field2', 'field3'])
+        new_df = pd.DataFrame(columns=['field1', 'field2'])
+        fields = ['field1', 'field2', 'field3']
+
+        updater_mock = mocker.Mock()
+        updater_mock.new_dataframe = new_df
+
+        with pytest.raises(RuntimeError) as exc_info:
+            palletjack.FeatureServiceInlineUpdater._validate_working_fields_in_live_and_new_dataframes(
+                updater_mock, live_df, fields
+            )
+        assert exc_info.value.args[
+            0
+        ] == 'Field mismatch between defined fields and either new or live data.\nFields not in live data: set()\nFields not in new data: {\'field3\'}'
+
+    def test_validate_working_fields_in_live_and_new_dataframes_raises_not_in_both(self, mocker):
+        live_df = pd.DataFrame(columns=['field1', 'field2'])
+        new_df = pd.DataFrame(columns=['field1', 'field2'])
+        fields = ['field1', 'field2', 'field3']
+
+        updater_mock = mocker.Mock()
+        updater_mock.new_dataframe = new_df
+
+        with pytest.raises(RuntimeError) as exc_info:
+            palletjack.FeatureServiceInlineUpdater._validate_working_fields_in_live_and_new_dataframes(
+                updater_mock, live_df, fields
+            )
+        assert exc_info.value.args[
+            0
+        ] == 'Field mismatch between defined fields and either new or live data.\nFields not in live data: {\'field3\'}\nFields not in new data: {\'field3\'}'
+
+
 class TestFeatureServiceInlineUpdaterIntegrated:
 
     def test_update_existing_features_in_hosted_feature_layer_no_matching_rows_returns_0(self, mocker, caplog):
@@ -1266,6 +1331,31 @@ class TestAttachments:
 
         tm.assert_frame_equal(current_attachments_df, test_df)
 
+    def test_check_attachment_dataframe_for_invalid_column_names_doesnt_raise_with_valid_names(self, mocker):
+        dataframe = pd.DataFrame(columns=['foo', 'bar'])
+        invalid_names = ['baz', 'boo']
+        palletjack.FeatureServiceAttachmentsUpdater._check_attachment_dataframe_for_invalid_column_names(
+            dataframe, invalid_names
+        )
+
+    def test_check_attachment_dataframe_for_invalid_column_names_raises_with_one_invalid(self, mocker):
+        dataframe = pd.DataFrame(columns=['foo', 'bar'])
+        invalid_names = ['foo', 'boo']
+        with pytest.raises(RuntimeError) as exc_info:
+            palletjack.FeatureServiceAttachmentsUpdater._check_attachment_dataframe_for_invalid_column_names(
+                dataframe, invalid_names
+            )
+        assert exc_info.value.args[0] == 'Attachment dataframe contains the following invalid names: [\'foo\']'
+
+    def test_check_attachment_dataframe_for_invalid_column_names_raises_with_all_invalid(self, mocker):
+        dataframe = pd.DataFrame(columns=['foo', 'bar'])
+        invalid_names = ['foo', 'bar']
+        with pytest.raises(RuntimeError) as exc_info:
+            palletjack.FeatureServiceAttachmentsUpdater._check_attachment_dataframe_for_invalid_column_names(
+                dataframe, invalid_names
+            )
+        assert exc_info.value.args[0] == 'Attachment dataframe contains the following invalid names: [\'foo\', \'bar\']'
+
     def test_add_attachments_by_oid_adds_and_doesnt_warn(self, mocker):
         action_df = pd.DataFrame({
             'OBJECTID': [1, 2],
@@ -1320,13 +1410,43 @@ class TestAttachments:
             },
         ]
 
-        updater_mock = mocker.Mock()
-        updater_mock.feature_layer.attachments.add.side_effect = result_dict
+        feature_layer_mock = mocker.Mock()
+        feature_layer_mock.attachments.add.side_effect = result_dict
+
+        updater = palletjack.FeatureServiceAttachmentsUpdater(mocker.Mock())
+        updater.feature_layer = feature_layer_mock
 
         with pytest.warns(UserWarning, match='Failed to attach path2 to OID 2'):
-            count = palletjack.FeatureServiceAttachmentsUpdater._add_attachments_by_oid(updater_mock, action_df, 'path')
+            count = updater._add_attachments_by_oid(action_df, 'path')
 
         assert count == 2
+        assert updater.failed_dict == {2: ('add', 'path2')}
+
+    def test_add_attachments_by_oid_handles_internal_agol_errors(self, mocker, caplog):
+        action_df = pd.DataFrame({
+            'OBJECTID': [1, 2],
+            'operation': ['add', 'add'],
+            'path': ['path1', 'path2'],
+        })
+
+        feature_layer_mock = mocker.Mock()
+        feature_layer_mock.attachments.add.side_effect = [
+            RuntimeError('foo'),
+            {
+                'addAttachmentResult': {
+                    'success': True
+                }
+            },
+        ]
+
+        updater = palletjack.FeatureServiceAttachmentsUpdater(mocker.Mock())
+        updater.feature_layer = feature_layer_mock
+
+        count = updater._add_attachments_by_oid(action_df, 'path')
+        assert count == 1
+        assert 'AGOL error while adding path1 to OID 1' in caplog.text
+        assert 'foo' in caplog.text
+        assert updater.failed_dict == {1: ('add', 'path1')}
 
     def test_add_attachments_by_oid_skips_overwrite_and_nan(self, mocker):
         action_df = pd.DataFrame({
@@ -1411,15 +1531,45 @@ class TestAttachments:
             },
         ]
 
-        updater_mock = mocker.Mock()
-        updater_mock.feature_layer.attachments.update.side_effect = result_dict
+        feature_layer_mock = mocker.Mock()
+        feature_layer_mock.attachments.update.side_effect = result_dict
+
+        updater = palletjack.FeatureServiceAttachmentsUpdater(mocker.Mock())
+        updater.feature_layer = feature_layer_mock
 
         with pytest.warns(UserWarning, match='Failed to update oldname2, attachment ID 22, on OID 2 with path2'):
-            count = palletjack.FeatureServiceAttachmentsUpdater._overwrite_attachments_by_oid(
-                updater_mock, action_df, 'path'
-            )
+            count = updater._overwrite_attachments_by_oid(action_df, 'path')
 
         assert count == 2
+        assert updater.failed_dict == {2: ('update', 'path2')}
+
+    def test_overwrite_attachments_by_oid_handles_internal_agol_errors(self, mocker, caplog):
+        action_df = pd.DataFrame({
+            'OBJECTID': [1, 2],
+            'operation': ['overwrite', 'overwrite'],
+            'path': ['path1', 'path2'],
+            'ID': [11, 22],
+            'NAME': ['oldname1', 'oldname2'],
+        })
+
+        feature_layer_mock = mocker.Mock()
+        feature_layer_mock.attachments.update.side_effect = [
+            RuntimeError('foo'),
+            {
+                'updateAttachmentResult': {
+                    'success': True
+                }
+            },
+        ]
+
+        updater = palletjack.FeatureServiceAttachmentsUpdater(mocker.Mock())
+        updater.feature_layer = feature_layer_mock
+
+        count = updater._overwrite_attachments_by_oid(action_df, 'path')
+        assert count == 1
+        assert 'AGOL error while overwritting oldname1 (attachment ID 11) on OID 1 with path1' in caplog.text
+        assert 'foo' in caplog.text
+        assert updater.failed_dict == {1: ('update', 'path1')}
 
     def test_overwrite_attachments_by_oid_skips_add_and_nan(self, mocker):
         action_df = pd.DataFrame({
@@ -1447,3 +1597,52 @@ class TestAttachments:
 
         assert updater_mock.feature_layer.attachments.update.call_count == 1
         assert count == 1
+
+    def test_create_attachments_dataframe_subsets_and_crafts_paths_properly(self, mocker):
+        input_df = pd.DataFrame({
+            'join': [1, 2, 3],
+            'pic': ['foo.png', 'bar.png', 'baz.png'],
+            'data': [11., 12., 13.],
+        })
+
+        attachment_df = palletjack.FeatureServiceAttachmentsUpdater.build_attachments_dataframe(
+            input_df, 'join', 'pic', '/foo/bar'
+        )
+
+        test_df = pd.DataFrame({
+            'join': [1, 2, 3],
+            'pic': ['foo.png', 'bar.png', 'baz.png'],
+            'full_file_path': [Path('/foo/bar/foo.png'),
+                               Path('/foo/bar/bar.png'),
+                               Path('/foo/bar/baz.png')]
+        })
+
+        #: Column of path objects won't test equal in assert_frame_equal, so we make lists of their str representations
+        #: and compare those separately from the rest of the dataframe
+        other_fields = ['join', 'pic']
+        tm.assert_frame_equal(attachment_df[other_fields], test_df[other_fields])
+        assert [str(path) for path in attachment_df['full_file_path']
+               ] == [str(path) for path in test_df['full_file_path']]
+
+    def test_create_attachments_dataframe_drops_missing_attachments(self, mocker):
+        input_df = pd.DataFrame({
+            'join': [1, 2, 3],
+            'pic': ['foo.png', None, ''],
+        })
+
+        attachment_df = palletjack.FeatureServiceAttachmentsUpdater.build_attachments_dataframe(
+            input_df, 'join', 'pic', '/foo/bar'
+        )
+
+        test_df = pd.DataFrame({
+            'join': [1],
+            'pic': ['foo.png'],
+            'full_file_path': [Path('/foo/bar/foo.png')],
+        })
+
+        #: Column of path objects won't test equal in assert_frame_equal, so we make lists of their str representations
+        #: and compare those separately from the rest of the dataframe
+        other_fields = ['join', 'pic']
+        tm.assert_frame_equal(attachment_df[other_fields], test_df[other_fields])
+        assert [str(path) for path in attachment_df['full_file_path']
+               ] == [str(path) for path in test_df['full_file_path']]
